@@ -32,7 +32,7 @@ fu_logitech_bulkcontroller_plugin_init(FuLogitechBulkcontrollerPlugin *self)
 static gboolean
 fu_logitech_bulkcontroller_plugin_write_firmware(FuPlugin *plugin,
 						 FuDevice *device,
-						 GInputStream *stream,
+						 FuFirmware *firmware,
 						 FuProgress *progress,
 						 FwupdInstallFlags flags,
 						 GError **error)
@@ -43,7 +43,7 @@ fu_logitech_bulkcontroller_plugin_write_firmware(FuPlugin *plugin,
 	locker = fu_device_locker_new(device, error);
 	if (locker == NULL)
 		return FALSE;
-	if (!fu_device_write_firmware(device, stream, progress, flags, error))
+	if (!fu_device_write_firmware(device, firmware, progress, flags, error))
 		return FALSE;
 	self->post_install = TRUE;
 	return TRUE;
@@ -65,6 +65,7 @@ static void
 fu_logitech_bulkcontroller_plugin_constructed(GObject *obj)
 {
 	FuPlugin *plugin = FU_PLUGIN(obj);
+	fu_plugin_add_udev_subsystem(plugin, "usb");
 	fu_plugin_set_device_gtype_default(plugin, FU_TYPE_LOGITECH_BULKCONTROLLER_DEVICE);
 	fu_plugin_add_device_gtype(plugin, FU_TYPE_LOGITECH_BULKCONTROLLER_CHILD); /* coverage */
 }
